==> dna-0.2.0.tar.gz: lib/library.dart <==
// Copyright (c) 2016, root ext. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

part of dna;

class DynamicLibrary<T> {
  final int STACK_SIZE = 256;
  static List<int> sizes = nGetSizes();

  Type type;
  String name;
  int libraryPointer;

  DynamicLibrary(){

    this.type = T;
    assert(type != dynamic);
    //print('type $type');

    InstanceMirror im = reflectClass(type).metadata.first.getField(#name);
    this.name = im.reflectee;

    libraryPointer = nLoadLibrary(this.name);
    //assert(libraryPointer != 0);
    //print('lib pointer $libraryPointer');
  }

  dynamic noSuchMethod(Invocation invocation){
    //print('method ${MirrorSystem.getName(invocation.memberName)}');

    var classMirror = reflectClass(type);
    var methods = classMirror.instanceMembers.values
        .where((e)=> e.isRegularMethod && e.owner == classMirror && e.simpleName == invocation.memberName).toList();
    assert(methods.isNotEmpty);
    var methodMirror = methods.first;

    //Array with TypedData object for params
    var paramsData = new List<TypedData>(invocation.positionalArguments.length);

    var methodPointer = nFindSymbol(libraryPointer, MirrorSystem.getName(invocation.memberName));
    assert(methodPointer != 0);
    //print('method pointer $methodPointer');

    var vm = nCreateCallVM(STACK_SIZE);
    assert(vm != 0);

    before(invocation, methodMirror, vm, paramsData);
    var result = call(methodMirror, vm, methodPointer);
    after(invocation, methodMirror, paramsData);

    nFreeCallVM(vm);
    return result;
  }

  void iterateParams(Invocation invocation, MethodMirror methodMirror,
    processParam(dynamic param, Type type, ParameterMirror paramMirror, int index)){

    var paramMirrors = methodMirror.parameters;

    int index = 0;
    for (var paramMirror in paramMirrors) {
      //print('param ${MirrorSystem.getName(param.simpleName)}');
      var argTag = Util.paramAnnot(paramMirror, Arg);
      var paramType = argTag != null ? argTag.getField(#type).reflectee : typeToArgType(paramMirror.type);
      var param = invocation.positionalArguments[index];

      processParam(param, paramType, paramMirror, index);
      index++;
    }
  }

  void before(Invocation invocation, MethodMirror methodMirror, int vm, List<TypedData> data) {
    var args = new List<Object>();
    var signature = new List<String>();

    iterateParams(invocation, methodMirror, (param, type, mirror, index){
      signature.add(new String.fromCharCodes([type]));
      args.add(param);

      switch (type){
        case ARG_POINTER: {

        }
        break;
        case ARG_STRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            String str = param.value;
            data[index] = new Uint8List(str.length);
            args[args.length-1] = data[index];
            signature[signature.length-1] = new String.fromCharCodes([ARG_TYPEDDATA]);
          }
        }
        break;
        case ARG_LISTINT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            List<int> list = param.value;
            data[index] = new Uint32List(list.length);
            args[args.length-1] = data[index];
            signature[signature.length-1] = new String.fromCharCodes([ARG_TYPEDDATA]);
          }
        }
        break;
        case ARG_LISTSTRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {

          }
        }
        break;
        case ARG_STRUCT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            data[index] = encode(param.value);
          } else {
            data[index] = encode(param);
          }

          args[args.length-1] = data[index];
          signature[signature.length-1] = new String.fromCharCodes([ARG_TYPEDDATA]);
        }
        break;
      }
    });

    nArgs(vm, signature.join(''), args);
  }

  void after(Invocation invocation, MethodMirror methodMirror, List<TypedData> data) {

    iterateParams(invocation, methodMirror, (param, type, mirror, index) {
      switch (type) {
        case ARG_POINTER: {

        }
        break;
        case ARG_STRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            Uint8List list = data[index];
            param.value = new String.fromCharCodes(list);
          }
        }
        break;
        case ARG_LISTINT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            Uint32List t = data[index];
            param.value = t.toList();
          }
        }
        break;
        case ARG_LISTSTRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {

          }
        }
        break;
        case ARG_STRUCT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            param.value = decode(data[index], param.value);
          }
        }
        break;
      }
    });
  }

  dynamic call(MethodMirror methodMirror, int vm, int methodPointer){

    var annot = Util.methodAnnot(methodMirror, Ret);
    var retType = annot != null ? annot.getField(#type).reflectee : typeToArgType(methodMirror.returnType);

    var result;
    if (retType == ARG_VOID){
      nCall(vm, methodPointer, new String.fromCharCodes([retType]));
    } else {
      result = nCall(vm, methodPointer, new String.fromCharCodes([retType]));
    }

    return result;
  }

  static int typeToArgType(Object type){
    if (type == reflectClass(int)){
      return ARG_INT;
    } else if (type == reflectClass(double)){
      return ARG_DOUBLE;
    } else if (type == reflectClass(bool)){
      return ARG_BOOL;
    } else if (type == reflectClass(String)){
      return ARG_STRING;
    } else {
      return ARG_VOID;
    }
  }

  static int argTypeToSize(int type){
    switch (type){
      case ARG_CHAR:
      case ARG_UCHAR:
        return sizes[0];
      case ARG_SHORT:
      case ARG_USHORT:
        return sizes[1];
      case ARG_INT:
      case ARG_UINT:
        return sizes[2];
      case ARG_LONG:
      case ARG_ULONG:
        return sizes[3];
      case ARG_LONGLONG:
      case ARG_ULONGLONG:
        return sizes[4];
      case ARG_BOOL:
        return sizes[5];
      case ARG_FLOAT:
        return sizes[6];
      case ARG_DOUBLE:
        return sizes[7];
      case ARG_POINTER:
        return sizes[8];
      default:
        return 0;
    }
  }

}

==> dna-0.2.1.tar.gz: lib/library.dart <==
// Copyright (c) 2016, root ext. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

part of dna;

class DynamicLibrary<T> {
  final int STACK_SIZE = 256;
  static List<int> sizes = getSizes();

  Type type;
  int libraryPointer;

  DynamicLibrary(){
    this.type = T;
    assert(type != dynamic);
    //print('type $type');

    var libTag = Util.classAnnot(reflectClass(type), Library);
    assert(libTag != null);
    var name = libTag.getField(#name).reflectee;
    var altName = libTag.getField(#altName).reflectee;

    libraryPointer = loadLibrary(name);
    if (libraryPointer == 0){
      libraryPointer = loadLibrary(altName);
    }
    //assert(libraryPointer != 0);
    //print('lib pointer $libraryPointer');
  }

  dynamic noSuchMethod(Invocation invocation){
    //print('method ${MirrorSystem.getName(invocation.memberName)}');

    var classMirror = reflectClass(type);
    var methods = classMirror.instanceMembers.values
        .where((e)=> e.isRegularMethod && e.owner == classMirror && e.simpleName == invocation.memberName).toList();
    assert(methods.isNotEmpty);
    var methodMirror = methods.first;

    //Array with TypedData object for params
    var paramsData = new List<TypedData>(invocation.positionalArguments.length);

    var methodPointer = findSymbol(libraryPointer, MirrorSystem.getName(invocation.memberName));
    assert(methodPointer != 0);
    //print('method pointer $methodPointer');

    var vm = createCallVM(STACK_SIZE);
    assert(vm != 0);

    before(invocation, methodMirror, vm, paramsData);
    var result = call(methodMirror, vm, methodPointer);
    after(invocation, methodMirror, paramsData);

    freeCallVM(vm);
    return result;
  }

  void iterateParams(Invocation invocation, MethodMirror methodMirror,
    processParam(dynamic param, Type type, ParameterMirror paramMirror, int index)){

    var paramMirrors = methodMirror.parameters;

    int index = 0;
    for (var paramMirror in paramMirrors) {
      //print('param ${MirrorSystem.getName(param.simpleName)}');
      var paramTag = Util.paramAnnot(paramMirror, Param);
      var paramType = paramTag != null ? paramTag.getField(#type).reflectee : dartTypeToCtype(paramMirror.type);
      var param = invocation.positionalArguments[index];

      processParam(param, paramType, paramMirror, index);
      index++;
    }
  }

  void before(Invocation invocation, MethodMirror methodMirror, int vm, List<TypedData> data) {
    var nativeParams = new List<Object>();
    var signature = new List<String>();

    iterateParams(invocation, methodMirror, (param, type, mirror, index){
      signature.add(new String.fromCharCodes([type]));
      nativeParams.add(param);

      switch (type){
        case C_POINTER: {

        }
        break;
        case C_STRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            String str = param.value;
            data[index] = new Uint8List(str.length);
            nativeParams[nativeParams.length-1] = data[index];
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          } else {
            nativeParams[nativeParams.length-1] = new Uint8List.fromList(param.codeUnits);
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          }
        }
        break;
        case LISTINT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            List<int> list = param.value;
            data[index] = new Uint32List(list.length);
            nativeParams[nativeParams.length-1] = data[index];
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          }
        }
        break;
        case LISTSTRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            throw new UnimplementedError();
          }
        }
        break;
        case C_STRUCT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            data[index] = encode(param.value);
          } else {
            data[index] = encode(param);
          }

          nativeParams[nativeParams.length-1] = data[index];
          signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
        }
        break;
      }
    });

    pushParameters(vm, signature.join(''), nativeParams);
  }

  void after(Invocation invocation, MethodMirror methodMirror, List<TypedData> data) {

    iterateParams(invocation, methodMirror, (param, type, mirror, index) {
      switch (type) {
        case C_POINTER: {

        }
        break;
        case C_STRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            Uint8List list = data[index];
            param.value = new String.fromCharCodes(list);
          }
        }
        break;
        case LISTINT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null){
            Uint32List t = data[index];
            param.value = t.toList();
          }
        }
        break;
        case LISTSTRING: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            throw new UnimplementedError();
          }
        }
        break;
        case C_STRUCT: {
          var outTag = Util.paramAnnot(mirror, Out);
          if (outTag != null) {
            param.value = decode(data[index], param.value);
          }
        }
        break;
      }
    });
  }

  dynamic call(MethodMirror methodMirror, int vm, int methodPointer){

    var retTag = Util.methodAnnot(methodMirror, Return);
    var retType = retTag != null ? retTag.getField(#type).reflectee : dartTypeToCtype(methodMirror.returnType);

    var result;
    if (retType == VOID){
      callFunction(vm, methodPointer, new String.fromCharCodes([retType]));
    } else {
      result = callFunction(vm, methodPointer, new String.fromCharCodes([retType]));
    }

    return result;
  }

  static int dartTypeToCtype(Object type){
    if (type == reflectClass(int)){
      return C_INT;
    } else if (type == reflectClass(double)){
      return C_DOUBLE;
    } else if (type == reflectClass(bool)){
      return C_BOOL;
    } else if (type == reflectClass(String)){
      return C_STRING;
    } else {
      return VOID;
    }
  }

  static int ctypeToSize(int type){
    switch (type){
      case C_CHAR:
      case C_UCHAR:
        return sizes[0];
      case C_SHORT:
      case C_USHORT:
        return sizes[1];
      case C_INT:
      case C_UINT:
        return sizes[2];
      case C_LONG:
      case C_ULONG:
        return sizes[3];
      case C_LONGLONG:
      case C_ULONGLONG:
        return sizes[4];
      case C_BOOL:
        return sizes[5];
      case C_FLOAT:
        return sizes[6];
      case C_DOUBLE:
        return sizes[7];
      case C_POINTER:
        return sizes[8];
      default:
        return 0;
    }
  }
}
